/* Do not edit this file. It is produced from the corresponding .m4 source */
/*
 *  Copyright (C) 2003, Northwestern University and Argonne National Laboratory
 *  See COPYRIGHT notice in top-level directory.
 */
/* $Id: bput.m4 2686 2016-12-07 20:02:41Z wkliao $ */

#if HAVE_CONFIG_H
# include <ncconfig.h>
#endif

#include <stdio.h>
#include <unistd.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#include <assert.h>

#include <mpi.h>

#include "nc.h"
#include "ncx.h"
#include "ncmpidtype.h"
#include "macro.h"

/*----< ncmpi_buffer_attach() >-----------------------------------------------*/
int
ncmpi_buffer_attach(int        ncid,
                    MPI_Offset bufsize)
{
    int err;
    NC *ncp;

    if (bufsize <= 0) DEBUG_RETURN_ERROR(NC_ENULLBUF)

    err = ncmpii_NC_check_id(ncid, &ncp);
    if (err != NC_NOERR) DEBUG_RETURN_ERROR(err)

    /* check if the buffer has been previously attached
     * note that in nc.c, the NC object is allocated with calloc, so
     * abuf should be initialized to NULL then
     */
    if (ncp->abuf != NULL) DEBUG_RETURN_ERROR(NC_EPREVATTACHBUF)

    ncp->abuf = (NC_buf*) NCI_Malloc(sizeof(NC_buf));

    ncp->abuf->size_allocated = bufsize;
    ncp->abuf->size_used = 0;
    ncp->abuf->table_size = NC_ABUF_DEFAULT_TABLE_SIZE;
    ncp->abuf->occupy_table = (NC_buf_status*)
               NCI_Calloc(NC_ABUF_DEFAULT_TABLE_SIZE, sizeof(NC_buf_status));
    ncp->abuf->tail = 0;
    ncp->abuf->buf = NCI_Malloc((size_t)bufsize);
    return NC_NOERR;
}

/*----< ncmpi_buffer_detach() >-----------------------------------------------*/
int
ncmpi_buffer_detach(int ncid)
{
    int  i, err;
    NC  *ncp;

    err = ncmpii_NC_check_id(ncid, &ncp);
    if (err != NC_NOERR) DEBUG_RETURN_ERROR(err)

    /* check if the buffer has been previously attached */
    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    /* this API assumes users are responsible for no pending bput */
    for (i=0; i<ncp->numPutReqs; i++) {
        if (ncp->put_list[i].abuf_index >= 0) /* check for a pending bput */
            DEBUG_RETURN_ERROR(NC_EPENDINGBPUT)
            /* return now, so users can call wait and try detach again */
    }

    NCI_Free(ncp->abuf->buf);
    NCI_Free(ncp->abuf->occupy_table);
    NCI_Free(ncp->abuf);
    ncp->abuf = NULL;

    return NC_NOERR;
}

#ifdef THIS_SEEMS_OVER_DONE_IT
/*----< ncmpi_buffer_detach() >-----------------------------------------------*/
/* mimic MPI_Buffer_detach()
 * Note from MPI: Even though the 'bufferptr' argument is declared as
 * 'void *', it is really the address of a void pointer.
 */
int
ncmpi_buffer_detach(int         ncid,
                    void       *bufptr,
                    MPI_Offset *bufsize)
{
    int  i, err;
    NC  *ncp;

    err = ncmpii_NC_check_id(ncid, &ncp);
    if (err != NC_NOERR) DEBUG_RETURN_ERROR(err)

    /* check if the buffer has been previously attached */
    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    /* check MPICH2 src/mpi/pt2pt/bsendutil.c for why the bufptr is void* */
    *(void **)bufptr = ncp->abuf->buf;
    *bufsize         = ncp->abuf->size_allocated;

    /* this API assumes users are responsible for no pending bput when called */
    for (i=0; i<ncp->numPutReqs; i++) {
        if (ncp->put_list[i].abuf_index >= 0) /* check for a pending bput */
            DEBUG_RETURN_ERROR(NC_EPENDINGBPUT)
            /* return now, so users can call wait and try detach again */
    }

    NCI_Free(ncp->abuf->occupy_table);
    NCI_Free(ncp->abuf);
    ncp->abuf = NULL;

    return NC_NOERR;
}
#endif


/*----< ncmpi_inq_buffer_usage() >--------------------------------------------*/
int
ncmpi_inq_buffer_usage(int         ncid,
                       MPI_Offset *usage) /* OUT: in bytes */
{
    int  err;
    NC  *ncp;

    err = ncmpii_NC_check_id(ncid, &ncp);
    if (err != NC_NOERR) DEBUG_RETURN_ERROR(err)

    /* check if the buffer has been previously attached */
    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    /* return the current usage in bytes */
    *usage = ncp->abuf->size_used;

    return NC_NOERR;
}

/*----< ncmpi_inq_buffer_size() >---------------------------------------------*/
int
ncmpi_inq_buffer_size(int         ncid,
                      MPI_Offset *buf_size) /* OUT: in bytes */
{
    int  err;
    NC  *ncp;

    err = ncmpii_NC_check_id(ncid, &ncp);
    if (err != NC_NOERR) DEBUG_RETURN_ERROR(err)

    /* check if the buffer has been previously attached */
    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    /* return the current usage in bytes */
    *buf_size = ncp->abuf->size_allocated;

    return NC_NOERR;
}


/*---- PnetCDF flexible APIs ------------------------------------------------*/
/*----< ncmpi_bput_var() >------------------------------------------------*/
int
ncmpi_bput_var(int ncid, int varid,  const void *buf,
                             MPI_Offset   bufcount,
                             MPI_Datatype buftype, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *start, *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, NULL,  NULL, NULL,
                              bufcount,
                              buftype,
                              API_VAR, 1,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_FULL_DIMENSIONS(start, count)

    /* ncmpi_bput_var is a special case of ncmpi_bput_varm */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              bufcount, buftype,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(start);
    return err;
}

/*----< ncmpi_bput_var1() >------------------------------------------------*/
int
ncmpi_bput_var1(int ncid, int varid, const MPI_Offset *start, const void *buf,
                             MPI_Offset   bufcount,
                             MPI_Datatype buftype, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, NULL, NULL,
                              bufcount,
                              buftype,
                              API_VAR1, 1,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_ONE_COUNT(count)

    /* ncmpi_bput_var1 is a special case of ncmpi_bput_varm */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              bufcount, buftype,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(count);
    return err;
}

/*----< ncmpi_bput_vara() >------------------------------------------------*/
int
ncmpi_bput_vara(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count, const void *buf,
                             MPI_Offset   bufcount,
                             MPI_Datatype buftype, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, NULL,
                              bufcount,
                              buftype,
                              API_VARA, 1,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vara is a special case of ncmpi_bput_varm */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              bufcount, buftype,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vars() >------------------------------------------------*/
int
ncmpi_bput_vars(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride, const void *buf,
                             MPI_Offset   bufcount,
                             MPI_Datatype buftype, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              bufcount,
                              buftype,
                              API_VARS, 1,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vars is a special case of ncmpi_bput_varm */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, NULL,
                              (void*)buf,
                              bufcount, buftype,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_varm() >------------------------------------------------*/
int
ncmpi_bput_varm(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride,
                   const MPI_Offset *imap, const void *buf,
                             MPI_Offset   bufcount,
                             MPI_Datatype buftype, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              bufcount,
                              buftype,
                              API_VARM, 1,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_varm is a special case of ncmpi_bput_varm */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, imap,
                              (void*)buf,
                              bufcount, buftype,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}



/*---- PnetCDF high-level APIs ----------------------------------------------*/
/*----< ncmpi_bput_var_text() >------------------------------------------------*/
int
ncmpi_bput_var_text(int ncid, int varid,  const char *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *start, *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, NULL,  NULL, NULL,
                              0,
                              MPI_CHAR,
                              API_VAR, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_FULL_DIMENSIONS(start, count)

    /* ncmpi_bput_var_text is a special case of ncmpi_bput_varm_text */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_CHAR,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(start);
    return err;
}

/*----< ncmpi_bput_var_schar() >------------------------------------------------*/
int
ncmpi_bput_var_schar(int ncid, int varid,  const schar *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *start, *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, NULL,  NULL, NULL,
                              0,
                              MPI_SIGNED_CHAR,
                              API_VAR, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_FULL_DIMENSIONS(start, count)

    /* ncmpi_bput_var_schar is a special case of ncmpi_bput_varm_schar */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_SIGNED_CHAR,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(start);
    return err;
}

/*----< ncmpi_bput_var_uchar() >------------------------------------------------*/
int
ncmpi_bput_var_uchar(int ncid, int varid,  const uchar *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *start, *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, NULL,  NULL, NULL,
                              0,
                              MPI_UNSIGNED_CHAR,
                              API_VAR, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_FULL_DIMENSIONS(start, count)

    /* ncmpi_bput_var_uchar is a special case of ncmpi_bput_varm_uchar */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_UNSIGNED_CHAR,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(start);
    return err;
}

/*----< ncmpi_bput_var_short() >------------------------------------------------*/
int
ncmpi_bput_var_short(int ncid, int varid,  const short *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *start, *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, NULL,  NULL, NULL,
                              0,
                              MPI_SHORT,
                              API_VAR, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_FULL_DIMENSIONS(start, count)

    /* ncmpi_bput_var_short is a special case of ncmpi_bput_varm_short */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_SHORT,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(start);
    return err;
}

/*----< ncmpi_bput_var_ushort() >------------------------------------------------*/
int
ncmpi_bput_var_ushort(int ncid, int varid,  const ushort *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *start, *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, NULL,  NULL, NULL,
                              0,
                              MPI_UNSIGNED_SHORT,
                              API_VAR, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_FULL_DIMENSIONS(start, count)

    /* ncmpi_bput_var_ushort is a special case of ncmpi_bput_varm_ushort */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_UNSIGNED_SHORT,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(start);
    return err;
}

/*----< ncmpi_bput_var_int() >------------------------------------------------*/
int
ncmpi_bput_var_int(int ncid, int varid,  const int *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *start, *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, NULL,  NULL, NULL,
                              0,
                              MPI_INT,
                              API_VAR, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_FULL_DIMENSIONS(start, count)

    /* ncmpi_bput_var_int is a special case of ncmpi_bput_varm_int */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_INT,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(start);
    return err;
}

/*----< ncmpi_bput_var_uint() >------------------------------------------------*/
int
ncmpi_bput_var_uint(int ncid, int varid,  const uint *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *start, *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, NULL,  NULL, NULL,
                              0,
                              MPI_UNSIGNED,
                              API_VAR, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_FULL_DIMENSIONS(start, count)

    /* ncmpi_bput_var_uint is a special case of ncmpi_bput_varm_uint */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_UNSIGNED,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(start);
    return err;
}

/*----< ncmpi_bput_var_long() >------------------------------------------------*/
int
ncmpi_bput_var_long(int ncid, int varid,  const long *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *start, *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, NULL,  NULL, NULL,
                              0,
                              MPI_LONG,
                              API_VAR, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_FULL_DIMENSIONS(start, count)

    /* ncmpi_bput_var_long is a special case of ncmpi_bput_varm_long */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_LONG,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(start);
    return err;
}

/*----< ncmpi_bput_var_float() >------------------------------------------------*/
int
ncmpi_bput_var_float(int ncid, int varid,  const float *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *start, *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, NULL,  NULL, NULL,
                              0,
                              MPI_FLOAT,
                              API_VAR, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_FULL_DIMENSIONS(start, count)

    /* ncmpi_bput_var_float is a special case of ncmpi_bput_varm_float */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_FLOAT,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(start);
    return err;
}

/*----< ncmpi_bput_var_double() >------------------------------------------------*/
int
ncmpi_bput_var_double(int ncid, int varid,  const double *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *start, *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, NULL,  NULL, NULL,
                              0,
                              MPI_DOUBLE,
                              API_VAR, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_FULL_DIMENSIONS(start, count)

    /* ncmpi_bput_var_double is a special case of ncmpi_bput_varm_double */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_DOUBLE,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(start);
    return err;
}

/*----< ncmpi_bput_var_longlong() >------------------------------------------------*/
int
ncmpi_bput_var_longlong(int ncid, int varid,  const long long *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *start, *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, NULL,  NULL, NULL,
                              0,
                              MPI_LONG_LONG_INT,
                              API_VAR, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_FULL_DIMENSIONS(start, count)

    /* ncmpi_bput_var_longlong is a special case of ncmpi_bput_varm_longlong */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_LONG_LONG_INT,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(start);
    return err;
}

/*----< ncmpi_bput_var_ulonglong() >------------------------------------------------*/
int
ncmpi_bput_var_ulonglong(int ncid, int varid,  const unsigned long long *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *start, *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, NULL,  NULL, NULL,
                              0,
                              MPI_UNSIGNED_LONG_LONG,
                              API_VAR, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_FULL_DIMENSIONS(start, count)

    /* ncmpi_bput_var_ulonglong is a special case of ncmpi_bput_varm_ulonglong */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_UNSIGNED_LONG_LONG,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(start);
    return err;
}

/*----< ncmpi_bput_var1_text() >------------------------------------------------*/
int
ncmpi_bput_var1_text(int ncid, int varid, const MPI_Offset *start, const char *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, NULL, NULL,
                              0,
                              MPI_CHAR,
                              API_VAR1, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_ONE_COUNT(count)

    /* ncmpi_bput_var1_text is a special case of ncmpi_bput_varm_text */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_CHAR,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(count);
    return err;
}

/*----< ncmpi_bput_var1_schar() >------------------------------------------------*/
int
ncmpi_bput_var1_schar(int ncid, int varid, const MPI_Offset *start, const schar *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, NULL, NULL,
                              0,
                              MPI_SIGNED_CHAR,
                              API_VAR1, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_ONE_COUNT(count)

    /* ncmpi_bput_var1_schar is a special case of ncmpi_bput_varm_schar */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_SIGNED_CHAR,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(count);
    return err;
}

/*----< ncmpi_bput_var1_uchar() >------------------------------------------------*/
int
ncmpi_bput_var1_uchar(int ncid, int varid, const MPI_Offset *start, const uchar *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, NULL, NULL,
                              0,
                              MPI_UNSIGNED_CHAR,
                              API_VAR1, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_ONE_COUNT(count)

    /* ncmpi_bput_var1_uchar is a special case of ncmpi_bput_varm_uchar */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_UNSIGNED_CHAR,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(count);
    return err;
}

/*----< ncmpi_bput_var1_short() >------------------------------------------------*/
int
ncmpi_bput_var1_short(int ncid, int varid, const MPI_Offset *start, const short *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, NULL, NULL,
                              0,
                              MPI_SHORT,
                              API_VAR1, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_ONE_COUNT(count)

    /* ncmpi_bput_var1_short is a special case of ncmpi_bput_varm_short */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_SHORT,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(count);
    return err;
}

/*----< ncmpi_bput_var1_ushort() >------------------------------------------------*/
int
ncmpi_bput_var1_ushort(int ncid, int varid, const MPI_Offset *start, const ushort *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, NULL, NULL,
                              0,
                              MPI_UNSIGNED_SHORT,
                              API_VAR1, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_ONE_COUNT(count)

    /* ncmpi_bput_var1_ushort is a special case of ncmpi_bput_varm_ushort */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_UNSIGNED_SHORT,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(count);
    return err;
}

/*----< ncmpi_bput_var1_int() >------------------------------------------------*/
int
ncmpi_bput_var1_int(int ncid, int varid, const MPI_Offset *start, const int *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, NULL, NULL,
                              0,
                              MPI_INT,
                              API_VAR1, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_ONE_COUNT(count)

    /* ncmpi_bput_var1_int is a special case of ncmpi_bput_varm_int */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_INT,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(count);
    return err;
}

/*----< ncmpi_bput_var1_uint() >------------------------------------------------*/
int
ncmpi_bput_var1_uint(int ncid, int varid, const MPI_Offset *start, const uint *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, NULL, NULL,
                              0,
                              MPI_UNSIGNED,
                              API_VAR1, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_ONE_COUNT(count)

    /* ncmpi_bput_var1_uint is a special case of ncmpi_bput_varm_uint */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_UNSIGNED,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(count);
    return err;
}

/*----< ncmpi_bput_var1_long() >------------------------------------------------*/
int
ncmpi_bput_var1_long(int ncid, int varid, const MPI_Offset *start, const long *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, NULL, NULL,
                              0,
                              MPI_LONG,
                              API_VAR1, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_ONE_COUNT(count)

    /* ncmpi_bput_var1_long is a special case of ncmpi_bput_varm_long */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_LONG,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(count);
    return err;
}

/*----< ncmpi_bput_var1_float() >------------------------------------------------*/
int
ncmpi_bput_var1_float(int ncid, int varid, const MPI_Offset *start, const float *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, NULL, NULL,
                              0,
                              MPI_FLOAT,
                              API_VAR1, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_ONE_COUNT(count)

    /* ncmpi_bput_var1_float is a special case of ncmpi_bput_varm_float */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_FLOAT,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(count);
    return err;
}

/*----< ncmpi_bput_var1_double() >------------------------------------------------*/
int
ncmpi_bput_var1_double(int ncid, int varid, const MPI_Offset *start, const double *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, NULL, NULL,
                              0,
                              MPI_DOUBLE,
                              API_VAR1, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_ONE_COUNT(count)

    /* ncmpi_bput_var1_double is a special case of ncmpi_bput_varm_double */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_DOUBLE,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(count);
    return err;
}

/*----< ncmpi_bput_var1_longlong() >------------------------------------------------*/
int
ncmpi_bput_var1_longlong(int ncid, int varid, const MPI_Offset *start, const long long *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, NULL, NULL,
                              0,
                              MPI_LONG_LONG_INT,
                              API_VAR1, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_ONE_COUNT(count)

    /* ncmpi_bput_var1_longlong is a special case of ncmpi_bput_varm_longlong */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_LONG_LONG_INT,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(count);
    return err;
}

/*----< ncmpi_bput_var1_ulonglong() >------------------------------------------------*/
int
ncmpi_bput_var1_ulonglong(int ncid, int varid, const MPI_Offset *start, const unsigned long long *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    MPI_Offset *count;

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, NULL, NULL,
                              0,
                              MPI_UNSIGNED_LONG_LONG,
                              API_VAR1, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    GET_ONE_COUNT(count)

    /* ncmpi_bput_var1_ulonglong is a special case of ncmpi_bput_varm_ulonglong */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_UNSIGNED_LONG_LONG,
                              reqid, WRITE_REQ, 1, 0);
    NCI_Free(count);
    return err;
}

/*----< ncmpi_bput_vara_text() >------------------------------------------------*/
int
ncmpi_bput_vara_text(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count, const char *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, NULL,
                              0,
                              MPI_CHAR,
                              API_VARA, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vara_text is a special case of ncmpi_bput_varm_text */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_CHAR,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vara_schar() >------------------------------------------------*/
int
ncmpi_bput_vara_schar(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count, const schar *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, NULL,
                              0,
                              MPI_SIGNED_CHAR,
                              API_VARA, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vara_schar is a special case of ncmpi_bput_varm_schar */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_SIGNED_CHAR,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vara_uchar() >------------------------------------------------*/
int
ncmpi_bput_vara_uchar(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count, const uchar *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, NULL,
                              0,
                              MPI_UNSIGNED_CHAR,
                              API_VARA, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vara_uchar is a special case of ncmpi_bput_varm_uchar */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_UNSIGNED_CHAR,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vara_short() >------------------------------------------------*/
int
ncmpi_bput_vara_short(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count, const short *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, NULL,
                              0,
                              MPI_SHORT,
                              API_VARA, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vara_short is a special case of ncmpi_bput_varm_short */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_SHORT,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vara_ushort() >------------------------------------------------*/
int
ncmpi_bput_vara_ushort(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count, const ushort *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, NULL,
                              0,
                              MPI_UNSIGNED_SHORT,
                              API_VARA, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vara_ushort is a special case of ncmpi_bput_varm_ushort */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_UNSIGNED_SHORT,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vara_int() >------------------------------------------------*/
int
ncmpi_bput_vara_int(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count, const int *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, NULL,
                              0,
                              MPI_INT,
                              API_VARA, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vara_int is a special case of ncmpi_bput_varm_int */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_INT,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vara_uint() >------------------------------------------------*/
int
ncmpi_bput_vara_uint(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count, const uint *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, NULL,
                              0,
                              MPI_UNSIGNED,
                              API_VARA, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vara_uint is a special case of ncmpi_bput_varm_uint */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_UNSIGNED,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vara_long() >------------------------------------------------*/
int
ncmpi_bput_vara_long(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count, const long *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, NULL,
                              0,
                              MPI_LONG,
                              API_VARA, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vara_long is a special case of ncmpi_bput_varm_long */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_LONG,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vara_float() >------------------------------------------------*/
int
ncmpi_bput_vara_float(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count, const float *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, NULL,
                              0,
                              MPI_FLOAT,
                              API_VARA, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vara_float is a special case of ncmpi_bput_varm_float */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_FLOAT,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vara_double() >------------------------------------------------*/
int
ncmpi_bput_vara_double(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count, const double *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, NULL,
                              0,
                              MPI_DOUBLE,
                              API_VARA, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vara_double is a special case of ncmpi_bput_varm_double */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_DOUBLE,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vara_longlong() >------------------------------------------------*/
int
ncmpi_bput_vara_longlong(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count, const long long *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, NULL,
                              0,
                              MPI_LONG_LONG_INT,
                              API_VARA, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vara_longlong is a special case of ncmpi_bput_varm_longlong */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_LONG_LONG_INT,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vara_ulonglong() >------------------------------------------------*/
int
ncmpi_bput_vara_ulonglong(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count, const unsigned long long *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, NULL,
                              0,
                              MPI_UNSIGNED_LONG_LONG,
                              API_VARA, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vara_ulonglong is a special case of ncmpi_bput_varm_ulonglong */
    err = ncmpii_igetput_varm(ncp, varp, start, count, NULL, NULL,
                              (void*)buf,
                              -1, MPI_UNSIGNED_LONG_LONG,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vars_text() >------------------------------------------------*/
int
ncmpi_bput_vars_text(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride, const char *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_CHAR,
                              API_VARS, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vars_text is a special case of ncmpi_bput_varm_text */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, NULL,
                              (void*)buf,
                              -1, MPI_CHAR,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vars_schar() >------------------------------------------------*/
int
ncmpi_bput_vars_schar(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride, const schar *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_SIGNED_CHAR,
                              API_VARS, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vars_schar is a special case of ncmpi_bput_varm_schar */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, NULL,
                              (void*)buf,
                              -1, MPI_SIGNED_CHAR,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vars_uchar() >------------------------------------------------*/
int
ncmpi_bput_vars_uchar(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride, const uchar *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_UNSIGNED_CHAR,
                              API_VARS, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vars_uchar is a special case of ncmpi_bput_varm_uchar */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, NULL,
                              (void*)buf,
                              -1, MPI_UNSIGNED_CHAR,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vars_short() >------------------------------------------------*/
int
ncmpi_bput_vars_short(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride, const short *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_SHORT,
                              API_VARS, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vars_short is a special case of ncmpi_bput_varm_short */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, NULL,
                              (void*)buf,
                              -1, MPI_SHORT,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vars_ushort() >------------------------------------------------*/
int
ncmpi_bput_vars_ushort(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride, const ushort *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_UNSIGNED_SHORT,
                              API_VARS, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vars_ushort is a special case of ncmpi_bput_varm_ushort */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, NULL,
                              (void*)buf,
                              -1, MPI_UNSIGNED_SHORT,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vars_int() >------------------------------------------------*/
int
ncmpi_bput_vars_int(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride, const int *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_INT,
                              API_VARS, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vars_int is a special case of ncmpi_bput_varm_int */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, NULL,
                              (void*)buf,
                              -1, MPI_INT,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vars_uint() >------------------------------------------------*/
int
ncmpi_bput_vars_uint(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride, const uint *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_UNSIGNED,
                              API_VARS, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vars_uint is a special case of ncmpi_bput_varm_uint */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, NULL,
                              (void*)buf,
                              -1, MPI_UNSIGNED,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vars_long() >------------------------------------------------*/
int
ncmpi_bput_vars_long(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride, const long *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_LONG,
                              API_VARS, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vars_long is a special case of ncmpi_bput_varm_long */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, NULL,
                              (void*)buf,
                              -1, MPI_LONG,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vars_float() >------------------------------------------------*/
int
ncmpi_bput_vars_float(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride, const float *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_FLOAT,
                              API_VARS, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vars_float is a special case of ncmpi_bput_varm_float */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, NULL,
                              (void*)buf,
                              -1, MPI_FLOAT,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vars_double() >------------------------------------------------*/
int
ncmpi_bput_vars_double(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride, const double *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_DOUBLE,
                              API_VARS, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vars_double is a special case of ncmpi_bput_varm_double */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, NULL,
                              (void*)buf,
                              -1, MPI_DOUBLE,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vars_longlong() >------------------------------------------------*/
int
ncmpi_bput_vars_longlong(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride, const long long *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_LONG_LONG_INT,
                              API_VARS, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vars_longlong is a special case of ncmpi_bput_varm_longlong */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, NULL,
                              (void*)buf,
                              -1, MPI_LONG_LONG_INT,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_vars_ulonglong() >------------------------------------------------*/
int
ncmpi_bput_vars_ulonglong(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride, const unsigned long long *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_UNSIGNED_LONG_LONG,
                              API_VARS, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_vars_ulonglong is a special case of ncmpi_bput_varm_ulonglong */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, NULL,
                              (void*)buf,
                              -1, MPI_UNSIGNED_LONG_LONG,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_varm_text() >------------------------------------------------*/
int
ncmpi_bput_varm_text(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride,
                   const MPI_Offset *imap, const char *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_CHAR,
                              API_VARM, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_varm_text is a special case of ncmpi_bput_varm_text */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, imap,
                              (void*)buf,
                              -1, MPI_CHAR,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_varm_schar() >------------------------------------------------*/
int
ncmpi_bput_varm_schar(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride,
                   const MPI_Offset *imap, const schar *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_SIGNED_CHAR,
                              API_VARM, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_varm_schar is a special case of ncmpi_bput_varm_schar */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, imap,
                              (void*)buf,
                              -1, MPI_SIGNED_CHAR,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_varm_uchar() >------------------------------------------------*/
int
ncmpi_bput_varm_uchar(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride,
                   const MPI_Offset *imap, const uchar *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_UNSIGNED_CHAR,
                              API_VARM, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_varm_uchar is a special case of ncmpi_bput_varm_uchar */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, imap,
                              (void*)buf,
                              -1, MPI_UNSIGNED_CHAR,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_varm_short() >------------------------------------------------*/
int
ncmpi_bput_varm_short(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride,
                   const MPI_Offset *imap, const short *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_SHORT,
                              API_VARM, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_varm_short is a special case of ncmpi_bput_varm_short */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, imap,
                              (void*)buf,
                              -1, MPI_SHORT,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_varm_ushort() >------------------------------------------------*/
int
ncmpi_bput_varm_ushort(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride,
                   const MPI_Offset *imap, const ushort *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_UNSIGNED_SHORT,
                              API_VARM, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_varm_ushort is a special case of ncmpi_bput_varm_ushort */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, imap,
                              (void*)buf,
                              -1, MPI_UNSIGNED_SHORT,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_varm_int() >------------------------------------------------*/
int
ncmpi_bput_varm_int(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride,
                   const MPI_Offset *imap, const int *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_INT,
                              API_VARM, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_varm_int is a special case of ncmpi_bput_varm_int */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, imap,
                              (void*)buf,
                              -1, MPI_INT,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_varm_uint() >------------------------------------------------*/
int
ncmpi_bput_varm_uint(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride,
                   const MPI_Offset *imap, const uint *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_UNSIGNED,
                              API_VARM, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_varm_uint is a special case of ncmpi_bput_varm_uint */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, imap,
                              (void*)buf,
                              -1, MPI_UNSIGNED,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_varm_long() >------------------------------------------------*/
int
ncmpi_bput_varm_long(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride,
                   const MPI_Offset *imap, const long *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_LONG,
                              API_VARM, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_varm_long is a special case of ncmpi_bput_varm_long */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, imap,
                              (void*)buf,
                              -1, MPI_LONG,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_varm_float() >------------------------------------------------*/
int
ncmpi_bput_varm_float(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride,
                   const MPI_Offset *imap, const float *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_FLOAT,
                              API_VARM, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_varm_float is a special case of ncmpi_bput_varm_float */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, imap,
                              (void*)buf,
                              -1, MPI_FLOAT,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_varm_double() >------------------------------------------------*/
int
ncmpi_bput_varm_double(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride,
                   const MPI_Offset *imap, const double *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_DOUBLE,
                              API_VARM, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_varm_double is a special case of ncmpi_bput_varm_double */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, imap,
                              (void*)buf,
                              -1, MPI_DOUBLE,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_varm_longlong() >------------------------------------------------*/
int
ncmpi_bput_varm_longlong(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride,
                   const MPI_Offset *imap, const long long *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_LONG_LONG_INT,
                              API_VARM, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_varm_longlong is a special case of ncmpi_bput_varm_longlong */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, imap,
                              (void*)buf,
                              -1, MPI_LONG_LONG_INT,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}

/*----< ncmpi_bput_varm_ulonglong() >------------------------------------------------*/
int
ncmpi_bput_varm_ulonglong(int ncid, int varid, const MPI_Offset *start,
                   const MPI_Offset *count,
                   const MPI_Offset *stride,
                   const MPI_Offset *imap, const unsigned long long *buf, int *reqid)
{
    int         err;
    NC         *ncp;
    NC_var     *varp=NULL;
    

    if (reqid != NULL) *reqid = NC_REQ_NULL;
    err = ncmpii_sanity_check(ncid, varid, start, count, stride,
                              0,
                              MPI_UNSIGNED_LONG_LONG,
                              API_VARM, 0,
                              0, WRITE_REQ, NONBLOCKING_IO, &ncp, &varp);
    if (err != NC_NOERR) return err;

    if (ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    

    /* ncmpi_bput_varm_ulonglong is a special case of ncmpi_bput_varm_ulonglong */
    err = ncmpii_igetput_varm(ncp, varp, start, count, stride, imap,
                              (void*)buf,
                              -1, MPI_UNSIGNED_LONG_LONG,
                              reqid, WRITE_REQ, 1, 0);
    
    return err;
}


