/* Do not edit this file. It is produced from the corresponding .m4 source */
/*
 *  Copyright (C) 2014, Northwestern University and Argonne National Laboratory
 *  See COPYRIGHT notice in top-level directory.
 */
/* $Id: i_varn.m4 2664 2016-12-02 00:50:43Z wkliao $ */

#if HAVE_CONFIG_H
# include <ncconfig.h>
#endif

#include <stdio.h>
#include <unistd.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#include <assert.h>

#include <mpi.h>

#include "nc.h"
#include "ncx.h"
#include "ncmpidtype.h"
#include "macro.h"





/* ncmpi_iget/iput_varn_<type>_<mode> API:
 *    type:   data type of I/O buffer, buf
 *    mode:   indpendent (<nond>) or collective (_all)
 *
 * arguments:
 *    num:    number of start and count pairs
 *    starts: an 2D array of size [num][ndims]. Each starts[i][*] indicates
 *            the starting array indices for a subarray request. ndims is
 *            the number of dimensions of the defined netCDF variable.
 *    counts: an 2D array of size [num][ndims]. Each counts[i][*] indicates
 *            the number of array elements to be accessed. This argument
 *            can be NULL, equivalent to counts with all 1s.
 *    bufcount and buftype: these 2 arguments are only available for flexible
 *            APIs, indicating the I/O buffer memory layout. When buftype is
 *            MPI_DATATYPE_NULL, bufcount is ignored and the data type of buf
 *            is considered matched the variable data type defined in the file.
 *    reqid:  request ID returned to user
 */

static int
ncmpii_igetput_varn(NC                *ncp,
                    NC_var            *varp,
                    int                num,
                    MPI_Offset* const *starts,  /* [num][varp->ndims] */
                    MPI_Offset* const *counts,  /* [num][varp->ndims] */
                    void              *buf,
                    MPI_Offset         bufcount,
                    MPI_Datatype       buftype,   /* data type of the bufer */
                    int               *reqid,
                    int                rw_flag,
                    int                use_abuf);


/*----< ncmpi_iput_varn() >-----------------------------------------------------*/
int
ncmpi_iput_varn(int                 ncid,
              int                 varid,
              int                 num,
              MPI_Offset* const  *starts,
              MPI_Offset* const  *counts,
              const void  *buf,
              MPI_Offset          bufcount,
              MPI_Datatype        buftype,
              int                *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero-size request */
    if (num == 0 || bufcount == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, bufcount,
                                 buftype, API_VARN, 1, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               bufcount, buftype, reqid, WRITE_REQ,
                               0);
}

/*----< ncmpi_iget_varn() >-----------------------------------------------------*/
int
ncmpi_iget_varn(int                 ncid,
              int                 varid,
              int                 num,
              MPI_Offset* const  *starts,
              MPI_Offset* const  *counts,
               void  *buf,
              MPI_Offset          bufcount,
              MPI_Datatype        buftype,
              int                *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero-size request */
    if (num == 0 || bufcount == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, bufcount,
                                 buftype, API_VARN, 1, 0, READ_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               bufcount, buftype, reqid, READ_REQ,
                               0);
}

/*----< ncmpi_bput_varn() >-----------------------------------------------------*/
int
ncmpi_bput_varn(int                 ncid,
              int                 varid,
              int                 num,
              MPI_Offset* const  *starts,
              MPI_Offset* const  *counts,
              const void  *buf,
              MPI_Offset          bufcount,
              MPI_Datatype        buftype,
              int                *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero-size request */
    if (num == 0 || bufcount == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, bufcount,
                                 buftype, API_VARN, 1, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               bufcount, buftype, reqid, WRITE_REQ,
                               1);
}


/*----< ncmpi_iput_varn_text() >--------------------------------------------------*/
int
ncmpi_iput_varn_text(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const char   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_CHAR, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_CHAR, reqid, WRITE_REQ, 0);
}

/*----< ncmpi_iput_varn_schar() >--------------------------------------------------*/
int
ncmpi_iput_varn_schar(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const schar   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_SIGNED_CHAR, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_SIGNED_CHAR, reqid, WRITE_REQ, 0);
}

/*----< ncmpi_iput_varn_uchar() >--------------------------------------------------*/
int
ncmpi_iput_varn_uchar(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const uchar   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_UNSIGNED_CHAR, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_UNSIGNED_CHAR, reqid, WRITE_REQ, 0);
}

/*----< ncmpi_iput_varn_short() >--------------------------------------------------*/
int
ncmpi_iput_varn_short(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const short   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_SHORT, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_SHORT, reqid, WRITE_REQ, 0);
}

/*----< ncmpi_iput_varn_ushort() >--------------------------------------------------*/
int
ncmpi_iput_varn_ushort(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const ushort   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_UNSIGNED_SHORT, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_UNSIGNED_SHORT, reqid, WRITE_REQ, 0);
}

/*----< ncmpi_iput_varn_int() >--------------------------------------------------*/
int
ncmpi_iput_varn_int(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const int   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_INT, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_INT, reqid, WRITE_REQ, 0);
}

/*----< ncmpi_iput_varn_uint() >--------------------------------------------------*/
int
ncmpi_iput_varn_uint(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const uint   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_UNSIGNED, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_UNSIGNED, reqid, WRITE_REQ, 0);
}

/*----< ncmpi_iput_varn_long() >--------------------------------------------------*/
int
ncmpi_iput_varn_long(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const long   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_LONG, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_LONG, reqid, WRITE_REQ, 0);
}

/*----< ncmpi_iput_varn_float() >--------------------------------------------------*/
int
ncmpi_iput_varn_float(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const float   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_FLOAT, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_FLOAT, reqid, WRITE_REQ, 0);
}

/*----< ncmpi_iput_varn_double() >--------------------------------------------------*/
int
ncmpi_iput_varn_double(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const double   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_DOUBLE, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_DOUBLE, reqid, WRITE_REQ, 0);
}

/*----< ncmpi_iput_varn_longlong() >--------------------------------------------------*/
int
ncmpi_iput_varn_longlong(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const long long   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_LONG_LONG_INT, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_LONG_LONG_INT, reqid, WRITE_REQ, 0);
}

/*----< ncmpi_iput_varn_ulonglong() >--------------------------------------------------*/
int
ncmpi_iput_varn_ulonglong(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const unsigned long long   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_UNSIGNED_LONG_LONG, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_UNSIGNED_LONG_LONG, reqid, WRITE_REQ, 0);
}

/*----< ncmpi_iget_varn_text() >--------------------------------------------------*/
int
ncmpi_iget_varn_text(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                  char   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_CHAR, API_VARN, 0, 0, READ_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_CHAR, reqid, READ_REQ, 0);
}

/*----< ncmpi_iget_varn_schar() >--------------------------------------------------*/
int
ncmpi_iget_varn_schar(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                  schar   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_SIGNED_CHAR, API_VARN, 0, 0, READ_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_SIGNED_CHAR, reqid, READ_REQ, 0);
}

/*----< ncmpi_iget_varn_uchar() >--------------------------------------------------*/
int
ncmpi_iget_varn_uchar(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                  uchar   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_UNSIGNED_CHAR, API_VARN, 0, 0, READ_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_UNSIGNED_CHAR, reqid, READ_REQ, 0);
}

/*----< ncmpi_iget_varn_short() >--------------------------------------------------*/
int
ncmpi_iget_varn_short(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                  short   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_SHORT, API_VARN, 0, 0, READ_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_SHORT, reqid, READ_REQ, 0);
}

/*----< ncmpi_iget_varn_ushort() >--------------------------------------------------*/
int
ncmpi_iget_varn_ushort(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                  ushort   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_UNSIGNED_SHORT, API_VARN, 0, 0, READ_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_UNSIGNED_SHORT, reqid, READ_REQ, 0);
}

/*----< ncmpi_iget_varn_int() >--------------------------------------------------*/
int
ncmpi_iget_varn_int(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                  int   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_INT, API_VARN, 0, 0, READ_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_INT, reqid, READ_REQ, 0);
}

/*----< ncmpi_iget_varn_uint() >--------------------------------------------------*/
int
ncmpi_iget_varn_uint(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                  uint   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_UNSIGNED, API_VARN, 0, 0, READ_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_UNSIGNED, reqid, READ_REQ, 0);
}

/*----< ncmpi_iget_varn_long() >--------------------------------------------------*/
int
ncmpi_iget_varn_long(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                  long   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_LONG, API_VARN, 0, 0, READ_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_LONG, reqid, READ_REQ, 0);
}

/*----< ncmpi_iget_varn_float() >--------------------------------------------------*/
int
ncmpi_iget_varn_float(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                  float   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_FLOAT, API_VARN, 0, 0, READ_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_FLOAT, reqid, READ_REQ, 0);
}

/*----< ncmpi_iget_varn_double() >--------------------------------------------------*/
int
ncmpi_iget_varn_double(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                  double   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_DOUBLE, API_VARN, 0, 0, READ_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_DOUBLE, reqid, READ_REQ, 0);
}

/*----< ncmpi_iget_varn_longlong() >--------------------------------------------------*/
int
ncmpi_iget_varn_longlong(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                  long long   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_LONG_LONG_INT, API_VARN, 0, 0, READ_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_LONG_LONG_INT, reqid, READ_REQ, 0);
}

/*----< ncmpi_iget_varn_ulonglong() >--------------------------------------------------*/
int
ncmpi_iget_varn_ulonglong(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                  unsigned long long   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_UNSIGNED_LONG_LONG, API_VARN, 0, 0, READ_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_UNSIGNED_LONG_LONG, reqid, READ_REQ, 0);
}

/*----< ncmpi_bput_varn_text() >--------------------------------------------------*/
int
ncmpi_bput_varn_text(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const char   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_CHAR, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_CHAR, reqid, WRITE_REQ, 1);
}

/*----< ncmpi_bput_varn_schar() >--------------------------------------------------*/
int
ncmpi_bput_varn_schar(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const schar   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_SIGNED_CHAR, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_SIGNED_CHAR, reqid, WRITE_REQ, 1);
}

/*----< ncmpi_bput_varn_uchar() >--------------------------------------------------*/
int
ncmpi_bput_varn_uchar(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const uchar   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_UNSIGNED_CHAR, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_UNSIGNED_CHAR, reqid, WRITE_REQ, 1);
}

/*----< ncmpi_bput_varn_short() >--------------------------------------------------*/
int
ncmpi_bput_varn_short(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const short   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_SHORT, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_SHORT, reqid, WRITE_REQ, 1);
}

/*----< ncmpi_bput_varn_ushort() >--------------------------------------------------*/
int
ncmpi_bput_varn_ushort(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const ushort   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_UNSIGNED_SHORT, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_UNSIGNED_SHORT, reqid, WRITE_REQ, 1);
}

/*----< ncmpi_bput_varn_int() >--------------------------------------------------*/
int
ncmpi_bput_varn_int(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const int   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_INT, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_INT, reqid, WRITE_REQ, 1);
}

/*----< ncmpi_bput_varn_uint() >--------------------------------------------------*/
int
ncmpi_bput_varn_uint(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const uint   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_UNSIGNED, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_UNSIGNED, reqid, WRITE_REQ, 1);
}

/*----< ncmpi_bput_varn_long() >--------------------------------------------------*/
int
ncmpi_bput_varn_long(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const long   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_LONG, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_LONG, reqid, WRITE_REQ, 1);
}

/*----< ncmpi_bput_varn_float() >--------------------------------------------------*/
int
ncmpi_bput_varn_float(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const float   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_FLOAT, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_FLOAT, reqid, WRITE_REQ, 1);
}

/*----< ncmpi_bput_varn_double() >--------------------------------------------------*/
int
ncmpi_bput_varn_double(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const double   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_DOUBLE, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_DOUBLE, reqid, WRITE_REQ, 1);
}

/*----< ncmpi_bput_varn_longlong() >--------------------------------------------------*/
int
ncmpi_bput_varn_longlong(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const long long   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_LONG_LONG_INT, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_LONG_LONG_INT, reqid, WRITE_REQ, 1);
}

/*----< ncmpi_bput_varn_ulonglong() >--------------------------------------------------*/
int
ncmpi_bput_varn_ulonglong(int                ncid,
                 int                varid,
                 int                num,
                 MPI_Offset* const *starts,
                 MPI_Offset* const *counts,
                 const unsigned long long   *buf,
                 int               *reqid)
{
    int     status;
    NC     *ncp;
    NC_var *varp=NULL;

    if (reqid != NULL) *reqid = NC_REQ_NULL;

    /* check for zero request */
    if (num == 0) return NC_NOERR;

    status = ncmpii_sanity_check(ncid, varid, NULL, NULL, NULL, 0,
                                 MPI_UNSIGNED_LONG_LONG, API_VARN, 0, 0, WRITE_REQ,
                                 NONBLOCKING_IO, &ncp, &varp);
    if (status != NC_NOERR) return status;

    /* set bufcount to -1 indicating non-flexible API */
    return ncmpii_igetput_varn(ncp, varp, num, starts, counts, (void*)buf,
                               -1, MPI_UNSIGNED_LONG_LONG, reqid, WRITE_REQ, 1);
}



/*----< ncmpii_igetput_varn() >-----------------------------------------------*/
static int
ncmpii_igetput_varn(NC                *ncp,
                    NC_var            *varp,
                    int                num,
                    MPI_Offset* const *starts,  /* [num][varp->ndims] */
                    MPI_Offset* const *counts,  /* [num][varp->ndims] */
                    void              *buf,
                    MPI_Offset         bufcount,
                    MPI_Datatype       buftype,   /* data type of the bufer */
                    int               *reqidp,    /* OUT: request ID */
                    int                rw_flag,   /* WRITE_REQ or READ_REQ */
                    int                use_abuf)  /* if use attached buffer */
{
    int i, j, el_size, status=NC_NOERR, free_cbuf=0, isSameGroup, reqid;
    void *cbuf=NULL;
    char *bufp;
    MPI_Offset **_counts=NULL;
    MPI_Datatype ptype;

    if (use_abuf && ncp->abuf == NULL) DEBUG_RETURN_ERROR(NC_ENULLABUF)

    /* it is illegal for starts to be NULL */
    if (starts == NULL)
        DEBUG_RETURN_ERROR(NC_ENULLSTART)
    else { /* it is illegal for any starts[i] to be NULL */
        for (i=0; i<num; i++) {
            if (starts[i] == NULL)
                DEBUG_RETURN_ERROR(NC_ENULLSTART)
        }
    }

    if (counts != NULL) {
        for (j=0; j<num; j++) {
            if (counts[j] == NULL)
                DEBUG_RETURN_ERROR(NC_ENULLCOUNT)
            for (i=0; i<varp->ndims; i++) {
                if (counts[j][i] < 0) /* no negative counts[][] */
                    DEBUG_RETURN_ERROR(NC_ENEGATIVECNT)
            }
        }
    }

    cbuf = buf;
    if (buftype == MPI_DATATYPE_NULL) {
        /* In this case, we make ptype match the variable data type defined
         * in file - no data conversion will be done. Also, it means buf is
         * contiguous. buftype will no longer be used.
         */
        ptype = ncmpii_nc2mpitype(varp->type);
        MPI_Type_size(ptype, &el_size); /* buffer element size */
    }
    else if (bufcount == -1) { /* if (IsPrimityMPIType(buftype)) */
        /* this subroutine is called from a high-level API
         * Also, it means the user buf is contiguous.
         * Assign ptype to buftype, and buftype will no longer be used.
         */
        ptype = buftype;
        MPI_Type_size(ptype, &el_size); /* buffer element size */
    }
    else { /* (bufcount > 0) flexible API is used */
        /* pack buf into cbuf, a contiguous buffer */
        int isderived, iscontig_of_ptypes;
        MPI_Offset bnelems=0;

        /* ptype (primitive MPI data type) from buftype
         * el_size is the element size of ptype
         * bnelems is the total number of ptype elements in buftype
         */
        status = ncmpii_dtype_decode(buftype, &ptype, &el_size, &bnelems,
                                     &isderived, &iscontig_of_ptypes);

        if (status != NC_NOERR) return status;

        if (bufcount != (int)bufcount) DEBUG_RETURN_ERROR(NC_EINTOVERFLOW)

        /* check if buftype is contiguous, if not, pack to one, cbuf */
        if (! iscontig_of_ptypes && bnelems > 0) {
            int position = 0;
            MPI_Offset packsize = bnelems*el_size;
            if (packsize != (int)packsize) DEBUG_RETURN_ERROR(NC_EINTOVERFLOW)

            cbuf = NCI_Malloc((size_t)packsize);
            free_cbuf = 1;
            /* if not use_abuf, need a callback to free cbuf */

            if (rw_flag == WRITE_REQ)
                MPI_Pack(buf, (int)bufcount, buftype, cbuf, (int)packsize,
                         &position, MPI_COMM_SELF);
        }
    }

    /* We allow counts == NULL and treat this the same as all 1s */
    if (counts == NULL) {
        _counts    = (MPI_Offset**) NCI_Malloc((size_t)num * sizeof(MPI_Offset*));
        _counts[0] = (MPI_Offset*)  NCI_Malloc((size_t)(num * varp->ndims *
                                                        SIZEOF_MPI_OFFSET));
        for (i=1; i<num; i++)
            _counts[i] = _counts[i-1] + varp->ndims;
        for (i=0; i<num; i++)
            for (j=0; j<varp->ndims; j++)
                _counts[i][j] = 1;
    }
    else {
        _counts = (MPI_Offset**) counts;
    }
    /* from this point forward, _counts != NULL */

    /* break buf into num pieces */
    reqid = NC_REQ_NULL;
    isSameGroup=0;
    bufp = (char*)cbuf;
    for (i=0; i<num; i++) {
        MPI_Offset buflen;

        /* check whether start, count, and stride are valid */
        status = NC_start_count_stride_ck(ncp, varp, starts[i], _counts[i], NULL, rw_flag);
        if (status != NC_NOERR) goto err_check;

        for (buflen=1, j=0; j<varp->ndims; j++)
            buflen *= _counts[i][j];

        if (buflen == 0) continue;
        status = ncmpii_igetput_varm(ncp, varp, starts[i], _counts[i], NULL,
                                     NULL, bufp, buflen, ptype, &reqid,
                                     rw_flag, use_abuf, isSameGroup);
        if (status != NC_NOERR) goto err_check;

        /* use isSamegroup so we end up with one nonblocking request (only the
         * first request gets a request ID back, the rest reuse the same ID.
         * This single ID represents num nonblocking requests */
        isSameGroup=1;
        bufp += buflen * el_size;
    }

    if (free_cbuf) { /* cbuf != buf, cbuf is temp allocated */
        if (rw_flag == READ_REQ) {
            /* Set the last lead request object to let wait() unpack cbuf to
             * buf and free cbuf */
            for (i=ncp->numGetReqs-1; i>=0; i--)
                if (ncp->get_list[i].num_recs > 0)
                    break;
            ncp->get_list[i].tmpBuf   = cbuf;
            ncp->get_list[i].userBuf  = buf;
            ncp->get_list[i].bufcount = (int)bufcount;
            MPI_Type_dup(buftype, &ncp->get_list[i].buftype);
        }
        else { /* WRITE_REQ */
            if (use_abuf)
                /* cbuf has been copied to the attached buffer, so it is safe
                 * to free cbuf now */
                NCI_Free(cbuf);
            else {
                for (i=ncp->numPutReqs-1; i>=0; i--)
                    if (ncp->put_list[i].num_recs > 0)
                        break;
                assert(i >= 0);
                /* Set the last request object to let wait() to free cbuf */
                ncp->put_list[i].tmpBuf = cbuf;
            }
        }
    }

err_check:
    if (_counts != counts) {
        NCI_Free(_counts[0]);
        NCI_Free(_counts);
    }

    if (status != NC_NOERR) {
        if (reqid != NC_REQ_NULL) /* cancel pending nonblocking request */
            ncmpii_cancel(ncp, 1, &reqid, NULL);
        if (free_cbuf) NCI_Free(cbuf);
    }
    if (reqidp != NULL) *reqidp = reqid;

    return status;
}
